"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function applyMutation(context, pipeline) {
    if (typeof context.mutate !== 'function') {
        return [context, pipeline];
    }
    var mutations = context.mutate(pipeline);
    var newPipeline = pipeline.slice();
    newPipeline.splice.apply(newPipeline, [context.mutationIndex, 0].concat(mutations.pipeline));
    context.mutate = undefined;
    return [context, newPipeline];
}
exports.applyMutation = applyMutation;
/**
 * @export
 * @template CallValue
 * @param {CallValue} callValue
 * @param {ExecutionContextFlow} executionFlow
 * @returns {IContext<CallValue>}
 */
function create(callValue, executionFlow) {
    return {
        callValue: callValue,
        executionFlow: executionFlow,
        mutationIndex: 0,
        previousValue: undefined,
        previousValues: [],
        call: function (callable) {
            return callable(this.previousValue, this);
        },
    };
}
exports.create = create;
/**
 *
 * @export
 * @template CallValue
 * @template PreviousValue
 * @param {IContext<CallValue>} context
 * @returns {IContext<CallValue, PreviousValue>}
 */
function downgrade(context) {
    var call = context.call, callValue = context.callValue, executionFlow = context.executionFlow, mutationIndex = context.mutationIndex, mutate = context.mutate, previousValues = context.previousValues;
    var previousMutationIndex = mutationIndex - 1;
    var prevPreviousValues = previousValues.slice(0, previousMutationIndex);
    return {
        call: call,
        callValue: callValue,
        executionFlow: executionFlow,
        mutationIndex: previousMutationIndex,
        mutate: mutate,
        previousValue: prevPreviousValues[prevPreviousValues.length - 1],
        previousValues: prevPreviousValues,
    };
}
exports.downgrade = downgrade;
/**
 * @export
 * @template CallValue
 * @template PreviousValue
 * @param {IContext<CallValue>} context
 * @param {PreviousValue} previousValue
 * @returns {IContext<CallValue, PreviousValue>}
 */
function update(context, previousValue) {
    var call = context.call, callValue = context.callValue, executionFlow = context.executionFlow, mutationIndex = context.mutationIndex, mutate = context.mutate, previousValues = context.previousValues;
    return {
        call: call,
        callValue: callValue,
        executionFlow: executionFlow,
        mutationIndex: mutationIndex + 1,
        mutate: mutate,
        previousValue: previousValue,
        previousValues: previousValues.concat([previousValue]),
    };
}
exports.update = update;
