"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var context_1 = require("./context");
function createFunction(pipeline, memoized) {
    var resolver = createResolver(pipeline, memoized);
    return function (arg) { return resolver(arg); };
}
function createIter(pipeline) {
    return function (callValue) {
        var context = context_1.create(callValue, 'sync');
        return createIterableMethods(callValue, pipeline, context);
    };
}
function createIterableMethods(callValue, pipeline, context) {
    var mutationIndex = context.mutationIndex;
    var methods = {};
    methods.hasnext = function () { return hasNextIterable(context, pipeline.length); };
    methods.hasprev = function () { return hasPrevIterable(context); };
    methods.next = function () {
        var _a = resolveSync(pipeline, mutationIndex, context, true), nextContext = _a[0], value = _a[1];
        return createIterableMethods(value, pipeline, nextContext);
    };
    methods.prev = function () {
        var prevContext = context_1.downgrade(context);
        return createIterableMethods(prevContext.previousValue, pipeline, prevContext);
    };
    methods.value = function () { return callValue; };
    return methods;
}
function createMethods(pipeline, memoized) {
    return {
        memoized: memoized,
        type: 'pipe',
        iter: createIter(pipeline),
        pipe: function (callable) {
            return createMethods(pipeline.concat([callable]), memoized);
        },
        resolve: createResolver(pipeline, memoized),
        resolveSync: createSyncResolver(pipeline, memoized),
        toFunction: function () { return createFunction(pipeline, memoized); },
        toSyncFunction: function () { return createSyncFunction(pipeline, memoized); },
    };
}
function createSyncFunction(pipeline, memoized) {
    var resolver = createSyncResolver(pipeline, memoized);
    return function (arg) { return resolver(arg); };
}
function createResolver(pipeline, memoized) {
    var _this = this;
    var previousCallValue;
    var previousReturnValue;
    return function (callValue) { return __awaiter(_this, void 0, void 0, function () {
        var _a, _, value;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    if (memoized && callValue === previousCallValue) {
                        return [2 /*return*/, previousReturnValue];
                    }
                    previousCallValue = callValue;
                    return [4 /*yield*/, resolve(pipeline, 0, context_1.create(callValue, 'async'))];
                case 1:
                    _a = _b.sent(), _ = _a[0], value = _a[1];
                    previousReturnValue = value;
                    return [2 /*return*/, previousReturnValue];
            }
        });
    }); };
}
function createSyncResolver(pipeline, memoized) {
    var previousCallValue;
    var previousReturnValue;
    return function (callValue) {
        if (memoized && callValue === previousCallValue) {
            return previousReturnValue;
        }
        previousCallValue = callValue;
        var _a = resolveSync(pipeline, 0, context_1.create(callValue, 'sync')), _ = _a[0], value = _a[1];
        previousReturnValue = value;
        return previousReturnValue;
    };
}
function hasNextIterable(context, pipelineSize) {
    return context.mutationIndex < pipelineSize;
}
function hasPrevIterable(context) {
    return context.mutationIndex > 0;
}
function resolve(pipeline, index, context, stopOnBreakpoint) {
    if (stopOnBreakpoint === void 0) { stopOnBreakpoint = false; }
    return __awaiter(this, void 0, void 0, function () {
        var previousValue, updatedContext, _a, mutatedContext, mutatedPipeline;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    if (index >= pipeline.length) {
                        return [2 /*return*/, [context, context.previousValue]];
                    }
                    return [4 /*yield*/, Promise.resolve(pipeline[index](context.previousValue, context))];
                case 1:
                    previousValue = _b.sent();
                    updatedContext = context_1.update(context, previousValue);
                    _a = context_1.applyMutation(updatedContext, pipeline), mutatedContext = _a[0], mutatedPipeline = _a[1];
                    return [2 /*return*/, resolve(mutatedPipeline, index + 1, mutatedContext, stopOnBreakpoint)];
            }
        });
    });
}
function resolveSync(pipeline, index, context, stopOnBreakpoint) {
    if (stopOnBreakpoint === void 0) { stopOnBreakpoint = false; }
    if (index >= pipeline.length) {
        return [context, context.previousValue];
    }
    var previousValue = pipeline[index](context.previousValue, context);
    var updatedContext = context_1.update(context, previousValue);
    var _a = context_1.applyMutation(updatedContext, pipeline), mutatedContext = _a[0], mutatedPipeline = _a[1];
    if (stopOnBreakpoint) {
        return [mutatedContext, previousValue];
    }
    return resolveSync(mutatedPipeline, index + 1, mutatedContext, stopOnBreakpoint);
}
/**
 * Creates a new pipeline
 * @export
 * @template CallValue
 * @template NextValue
 * @param {ExplicitCallable<CallValue, NextValue>} callable
 * @param {boolean} [memoized=false]
 * @returns
 */
function pipe(callable, memoized) {
    if (memoized === void 0) { memoized = false; }
    var pipeline = [callable];
    return createMethods(pipeline, memoized);
}
exports.pipe = pipe;
